﻿'www.vb-paradise.de/index.php/User/23082-exc-jdbi/
'
'github.com/exc-jdbi
'
'Is Free and Opensource

Option Strict On
Option Explicit On
Imports System.Text
Imports System.Numerics
Imports System.Security.Cryptography
Imports System.Runtime.InteropServices

''' <summary>
''' Modified RNGCryptoServiceProvider
''' </summary>
''' <history>© exc-jdbi 2017</history>
<System.Diagnostics.DebuggerStepThrough()> _
Public Class CryptoRnd
    Private _min As Int64
    Private _max As UInt64
    Private _maxf As Double = 0.0
    Private _minf As Double = Double.MaxValue
    Private rng As New RNGCryptoServiceProvider()
#Region "Secure Randoms"
    ''' <summary>
    ''' Set Secure-Random Area - Interval min / max. 
    ''' </summary>
    ''' <param name="min">Minimum</param>
    ''' <param name="max">Maximum</param>
    ''' <returns>Return True if min / max correctly</returns>
    Public Function Interval(Of T As IComparable)(ByVal min As T, ByVal max As T) As Boolean
        If max.CompareTo(min) > 0 Then
            Me._min = TryChangeType(Of Int64, T)(min)
            Me._max = TryChangeType(Of UInt64, T)(max)
            Return True
        End If
        Return False
    End Function
    ''' <summary>
    ''' Return a Array with Secure-Randoms as T
    ''' </summary>
    ''' <param name="arr">Return a Array with Secure-Randoms</param>
    Public Sub GetRnds(Of T As IComparable)(ByRef arr As T())
        If (arr IsNot Nothing) AndAlso (arr.Length > 0) Then
            For i As Int32 = 0 To arr.Length - 1
                arr(i) = Me.[Next](Of T)()
            Next
        End If
    End Sub
    ''' <summary>
    ''' Return a Array with Secure-Randoms  as T in Interval min / max. 
    ''' </summary>
    ''' <param name="min">Minimum</param>
    ''' <param name="max">Maximum</param>
    ''' <param name="arr">Return a Array with Secure-Randoms</param>
    Public Sub GetRnds(Of T As IComparable)(ByRef arr As T(), ByVal min As T, ByVal max As T)
        If (arr IsNot Nothing) AndAlso (arr.Length > 0) Then
            If Me.Interval(Of T)(min, max) Then
                Me.GetRnds(Of T)(arr)
            End If
        End If
    End Sub
    ''' <summary>
    ''' Next Secure-Random as T
    ''' </summary>  
    ''' <param name="min">Minimum</param>
    ''' <param name="max">Maximum</param>
    ''' <returns>Returns a Secure-Number as T</returns>
    Public Function [Next](Of T As IComparable)(ByVal min As T, ByVal max As T) As T
        If Me.Interval(Of T)(min, max) Then
            Return Me.[Next](Of T)()
        End If
        Return Nothing
    End Function
    ''' <summary>
    ''' Next Secure-Random as T
    ''' </summary>
    ''' <returns>Returns a Secure-Number as T</returns>
    Public Function [Next](Of T As IComparable)() As T
        Dim res As T = Nothing
        Dim decCount As Int32 = 0
        Dim n As BigInteger, r As Decimal

        'Calc Difference
        n = New BigInteger(Me._max) - New BigInteger(Me._min)
        r = CDec(Me.NextRnd())

        'Calculate the number of decimal places
        decCount = GetDecCount(r)

        'Accurate calculation with biginteger
        Dim expDec As UInt64 = CULng(10 ^ decCount)
        Dim tt As New BigInteger(r * expDec)

        'And divide again
        tt *= n : tt /= expDec

        Dim bi2 As New BigInteger(Me._min)
        If tt >= n Then
            res = TryChangeType2(Of T, UInt64)(Me._max)
        Else : res = TryChangeType2(Of T, Bignum)(New Bignum(bi2 + tt))
        End If
        Return res
    End Function
#End Region
#Region "Secure Random Primes"
    ''' <summary>
    ''' Next Secure-RandomPrime as T
    ''' </summary>
    ''' <returns>Returns a Secure-Prime as T</returns>
    Public Function NextPrime(Of T As IComparable)() As T
        If (_min >= 0) AndAlso (_max >= 0) Then
            Return Me.FindPrime(Of T)()
        End If
    End Function
    ''' <summary>
    ''' Next Secure-RandomPrime as T
    ''' </summary>
    ''' <param name="min">Minimum</param>
    ''' <param name="max">Maximum</param>
    ''' <returns>Returns a Secure-Prime as T</returns>
    Public Function NextPrime(Of T As IComparable)(ByVal min As T, ByVal max As T) As T
        If Me.Interval(Of T)(min, max) Then
            If (_min >= 0) AndAlso (_max >= 0) Then
                Return Me.FindPrime(Of T)()
            End If
        End If
    End Function
    ''' <summary>
    ''' Return a Array with Secure-RandomsPrimes as T
    ''' </summary>
    ''' <param name="arr">Return a Array with Secure-RandomsPrimes</param>
    Public Sub GetRndPrimes(Of T As IComparable)(ByRef arr As T())
        If (_min >= 0) AndAlso (_max >= 0) Then
            If (arr IsNot Nothing) AndAlso (arr.Length > 0) Then
                For i As Int32 = 0 To arr.Length - 1
                    arr(i) = Me.FindPrime(Of T)()
                Next
            End If
        End If
    End Sub
    ''' <summary>
    ''' Return a Array with Secure-RandomsPrimes as T in Interval min / max. 
    ''' </summary>
    ''' <param name="min">Minimum</param>
    ''' <param name="max">Maximum</param>
    ''' <param name="arr">Return a Array with Secure-RandomsPrimes</param>
    Public Sub GetRndPrimes(Of T As IComparable)(ByRef arr As T(), ByVal min As T, ByVal max As T)
        If Me.Interval(Of T)(min, max) Then
            Me.GetRndPrimes(Of T)(arr)
        End If
    End Sub
#End Region
#Region "Secure Random Double"
    ''' <summary>
    ''' Return a Secure-Randoms as Double less 1.0
    ''' </summary>
    Public Function NextDbl() As Double
        Return Me.NextRnd()
        'return this.GetRnd() * (1.0D / (double)(UInt64.MaxValue));
    End Function
    ''' <summary>
    ''' Return a Array with Secure-Randoms as Double less 1.0
    ''' </summary>
    ''' <param name="arr">Return a Array with Secure-Randoms</param>
    Public Sub GetRndsDbl(ByRef arr As Double())
        If (arr IsNot Nothing) AndAlso (arr.Length > 0) Then
            For i As Int32 = 0 To arr.Length - 1
                arr(i) = Me.NextDbl()
            Next
        End If
    End Sub
#End Region
#Region "Secure Random String"
    ''' <summary>
    ''' Returns a String with Secure-Randoms Length/Char alphanumeric
    ''' </summary>
    ''' <returns>Returns a Secure-RandomString alphanumeric</returns>
    Public Function GetRndAlpha() As String
        Dim max As Int32 = CInt(2 ^ 8)
        Dim len As Int32 = Me.Next(Of Int32)(8, max)
        Return GetRndAlpha(len)
    End Function
    ''' <summary>
    ''' Returns a String with Secure-Randoms as Char alphanumeric
    ''' </summary>
    ''' <param name="len">String.length</param>
    ''' <returns>Returns a Secure-RandomString alphanumeric</returns>
    Public Function GetRndAlpha(ByVal len As Int32) As String
        If len > 7 Then
            Dim res As New StringBuilder(len)
            Dim alpha As String = AlphaString()
            For i As Int32 = 0 To len - 1
                res.Append(alpha(Me.Next(0, alpha.Length - 1)))
            Next
            Return res.ToString
        End If
        Return String.Empty
    End Function
    ''' <summary>
    ''' Returns a String with Secure-Randoms Length/Char asciinumeric
    ''' </summary>
    ''' <returns>Returns a Secure-RandomString asciinumeric</returns>
    Public Function GetRndAscii() As String
        Dim max As Int32 = CInt(2 ^ 8)
        Dim len As Int32 = Me.Next(Of Int32)(8, max)
        Return GetRndAscii(len)
    End Function
    ''' <summary>
    ''' Returns a String with Secure-Randoms as Char asciinumeric
    ''' </summary>
    ''' <param name="len">String.length</param>
    ''' <returns>Returns a Secure-RandomString asciinumeric</returns>
    Public Function GetRndAscii(ByVal len As Int32) As String
        If len > 7 Then
            Dim res As New StringBuilder(len)
            For i As Int32 = 0 To len - 1
                res.Append(ChrW(Me.Next(Of Byte)(33, 255)))
            Next
            Return res.ToString
        End If
        Return String.Empty
    End Function
#End Region
#Region "Methodes"
    <System.Diagnostics.DebuggerHidden()> _
    Private Function AlphaString() As String
        Dim res As New StringBuilder(62)
        Dim z() As Byte = {48, 57, 65, 90, 97, 122}
        For i As Int32 = 0 To z.Length - 1 Step 2
            For j As Byte = z(i) To z(i + 1)
                res.Append(ChrW(j))
            Next
        Next
        If res.Length > 0 Then Return res.ToString
        Return String.Empty
    End Function
    <System.Diagnostics.DebuggerHidden()> _
    Private Function NextRnd() As Double
        Dim tmp As Double = Me.GetRnd() * CDbl(1.0 / CDbl(UInt64.MaxValue))
        If tmp < Me._minf Then Me._minf = tmp
        If tmp > Me._maxf Then Me._maxf = tmp
        If Me._maxf - Me._minf = 0.0 Then
            Me.NextRnd()
        End If
        Return 1.0 - ((Me._maxf - tmp) / (Me._maxf - Me._minf))
    End Function
    <System.Diagnostics.DebuggerHidden()> _
    Private Function GetRnd() As Double
        Dim scale As UInt64 = 0UL
        Dim size As Int32 = Marshal.SizeOf(GetType(UInt64))
        Dim buffer As Byte() = New Byte(size - 1) {}

        Me.rng.GetBytes(buffer)
        scale = BitConverter.ToUInt64(buffer, 0)
        Return CDbl(scale)
    End Function
    <System.Diagnostics.DebuggerHidden()> _
    Private Function NextUni() As UInt64
        Dim res As UInt64 = 0, n As Double, r As Double, t As Double
        n = CDbl(_max - _min) : r = NextRnd() : t = r * n
        Dim b As UInt64 = _max - CULng(_min)
        If t >= b Then
            res = _max
        Else : res = CULng(_min) + CULng(If(t > b, b, t))
        End If
        Return res
    End Function
    <System.Diagnostics.DebuggerHidden()> _
    Private Function FindPrime(Of T As IComparable)() As T
        FindPrime = Nothing
        Dim candidate As UInt64 = 0
        Dim bForwart As Boolean = True
        While True
            candidate = NextUni()
            If candidate < 3 Then Continue While
            If candidate Mod 2UL = 0UL Then candidate -= 1UL
            While True
                If (candidate >= _min) AndAlso (candidate <= _max) Then
                    If (candidate <= 5UL) OrElse ((candidate > 5UL) AndAlso (Not candidate Mod 5UL = 0UL)) Then
                        If BigInteger.ModPow(2, candidate, candidate) = 2 Then
                            If isPrimeMR(candidate) Then
                                Return TryChangeType2(Of T, UInt64)(candidate)
                            End If
                        End If
                    End If
                End If
                If (candidate < (_min + 2L)) Then bForwart = True
                If (candidate > (_max - 2UL)) Then bForwart = False
                If bForwart Then
                    candidate += 2UL
                Else : candidate -= 2UL
                End If
            End While
        End While
    End Function
    <System.Diagnostics.DebuggerHidden()> _
    Private Function isPrimeMR(ByVal candidate As BigInteger) As Boolean
        Dim s As Int32 = 0
        Dim d As BigInteger = candidate - 1
        'If (candidate < 0) OrElse ((candidate And 1) = 0) Then Return False
        While (d And 1) = 0
            d >>= 1 : s += 1
        End While

        Dim k As BigInteger
        For i As Int32 = 2 To 5
            k = BigInteger.ModPow(i, d, candidate)
            If (k = 1) OrElse (k = candidate - 1) Then Continue For
            For j As Int32 = 1 To s - 1
                k = BigInteger.ModPow(k, 2, candidate)
                If k = 1 Then Return False
                If k = candidate - 1 Then Exit For
            Next
            If Not k = candidate - 1 Then Return False
        Next
        Return True
    End Function
    <System.Diagnostics.DebuggerHidden()> _
    Private Function GetDecCount(ByVal dec As Decimal) As Int32
        Dim res As Int32 = ([Decimal].GetBits(dec)(3) And &HFF0000) >> &H10
        If res > 19 Then res = 19
        Return res
    End Function
    <System.Diagnostics.DebuggerHidden()> _
    Private Function TryChangeType(Of U As IConvertible, V)(ByVal val As V) As U
        Try
            Dim typ1 As Type = GetType(V), typ2 As Type = GetType(U)
            Return DirectCast(Convert.ChangeType(Convert.ChangeType(val, typ1), typ2), U)
        Catch generatedExceptionName As Exception
        End Try
        Return Nothing
    End Function
    <System.Diagnostics.DebuggerHidden()> _
    Private Function TryChangeType2(Of U, V As IConvertible)(ByVal val As V) As U
        Try
            Dim typ1 As Type = GetType(V), typ2 As Type = GetType(U)
            Return DirectCast(Convert.ChangeType(Convert.ChangeType(val, typ1), typ2), U)
        Catch generatedExceptionName As Exception
        End Try
        Return Nothing
    End Function
#End Region
#Region "De- and Contructor"
    <System.Diagnostics.DebuggerHidden()> _
    Public Sub New()
        Me._min = 0
        Me._max = 100
        Dim d As Double = NextRnd()
        For i As Int32 = 0 To 100
            d = Me.NextRnd()
        Next
    End Sub
    '<System.Diagnostics.DebuggerHidden()> _
    'Protected Overrides Sub Finalize()
    '    GC.Collect()
    '    MyBase.Finalize()
    '    me.rng.Dispose()
    'End Sub
#End Region
End Class
